﻿using FLY.Modbus;
using FLY.OBJComponents.IService;
using FLY.OBJComponents.Server;
using FLY.Weight.Common;
using FLY.Weight.IService;
using FLY.Weight.Server.Model;
using FObjBase;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;

namespace FLY.Weight.Server
{
    [JsonObject(MemberSerialization.OptIn)]
    public class WeightSystem : IWeightSystemService, IPropertyOpt
    {

        #region IWeightSystemService 接口

        public ObservableCollection<WeighterC> Items { get; } = new ObservableCollection<WeighterC>();

        /// <summary>
        /// 每1秒保存一次配料数据， 一共10秒。
        /// 当卸料时，提起3秒前的数据保存
        /// </summary>
        List<Dictionary<WeighterC, Lc_Mix>> mix10 = new List<Dictionary<WeighterC, Lc_Mix>>();

        public WeighterAccessory Accessory { get; } = new WeighterAccessory();

        private PLCProxySystem plcos = new PLCProxySystem();
        /// <summary>
        /// PLC代理系统
        /// </summary>
        public IPLCProxySystemService PLCos { get { return plcos; } }

        private int flowInterval = 10;
        /// <summary>
        /// 流量记录周期,单位s, 最小值为5
        /// </summary>
        [JsonProperty]
        public int FlowInterval
        {
            get { return flowInterval; }
            set
            {
                if (value < 5)
                    value = 5;
                if (flowInterval != value)
                {
                    flowInterval = value;
                }
            }
        }

        /// <summary>
        /// 层数
        /// </summary>
        public int ItemsCnt { get; private set; } = 3;

        /// <summary>
        /// 每层仓数
        /// </summary>
        public int[] BinCnts { get; private set; } = new int[] { 4, 4, 4 };
        #endregion


        /// <summary>
        /// 报警系统
        /// </summary>
        WarningSystem warning;
        /// <summary>
        /// 记录到数据库
        /// </summary>
        HistoryDb historyDb;
        /// <summary>
        /// 报警配置
        /// </summary>
        ErrorConf errorConf;

        public WeightSystem()
        {
            Load();

            //--------------------------------------------------------------------------------
            //step 1 加载PLC寄存器 文件
            AddConfigFile();



        }

        public void Init(HistoryDb historyDb, WarningSystem warning)
        {
            this.historyDb = historyDb;
            this.warning = warning;

            //--------------------------------------------------------------------------------
            //step 2 报警配置
            errorConf = new ErrorConf(PLCos, this.warning, "称重");
            errorConf.AddErrorAction(Accessory);
            for (int i = 0; i < Items.Count(); i++)
            {
                errorConf.AddErrorAction(
                    Items[i],
                        (ref string description, object state) => {
                            int idx = (int)state;
                            description = Items[idx].Number + "层 " + description;
                        },
                        i);
            }
            errorConf.InitError();



            Misc.BindingOperations.SetBinding(Items[0], "AlarmIsOn", () =>
            {
                this.warning.Enable = !Items[0].AlarmIsOn;
            });
            Misc.BindingOperations.SetBinding(this.warning, "Enable", () =>
            {
                for (int i = 0; i < ItemsCnt; i++)
                {
                    Items[i].AlarmIsOn = !this.warning.Enable;
                }
            });
            Misc.BindingOperations.SetBinding(this.warning, "IsRinging", () =>
            {
                if (!this.warning.IsRinging)
                {
                    for (int i = 0; i < ItemsCnt; i++)
                    {
                        Items[i].IsAlarmReseted = false;
                    }
                }
            });

            //--------------------------------------------------------------------------------
            //step 3 历史数据记录

            //--------------------------------------------------------------------------------
            //配料记录
            foreach (WeighterC w in Items)
            {
                //LastCumulativeProductions.Add(w, w.CumulativeProduction);
                w.PropertyChanged += WeightSystem_MBServer_PropertyChanged;
            }
            //--------------------------------------------------------------------------------
            //Test();


            this.PropertyChanged += (s, e) =>
            {
                if (e.PropertyName == "FlowInterval")
                {
                    Save();
                }
            };
            FObjBase.PollModule.Current.Poll_Config(PollModule.POLL_CONFIG.ADD, OnPoll, TimeSpan.FromSeconds(1));

            FObjBase.PollModule.Current.Poll_Config(PollModule.POLL_CONFIG.ADD, OnPoll_SaveMix, TimeSpan.FromSeconds(1));

            //--------------------------------------------------------------------------------
            //添加任务
            for (int i = 0; i < Items.Count(); i++)
            {
                List<string> props = new List<string>
                {
                    "BucketValveIsOpen",
                    "IsMixSave",
                    "MixDisp",
                    "ClearProduction",
                    "CumulativeProduction",
                    "CurrentFlow",
                    "ScrewPDisp",
                    "ScrewMotorFreq"
                };

                for (int j = 0; j < Items[i].BinCnt; j++)
                {
                    int no = j + 1;
                    props.Add($"MixSet_{no}");
                    props.Add($"MixPSet_{no}");
                    props.Add($"MixDisp_{no}");
                    props.Add($"MixPDisp_{no}");
                }

                PLCos.SetPlan($"Items[{i}]", props, 0);
            }
            PLCos.SetPlan("Accessory", new string[] { "TotalFlow" }, 0);

            //--------------------------------------------------------------------------------
            //last step  PLC 更新计划服务初始化
            plcos.Init();
        }
        void OnPoll_SaveMix()
        {
            Dictionary<WeighterC, Lc_Mix> mix = new Dictionary<WeighterC, Lc_Mix>();
            foreach (WeighterC w in Items)
            {
                //记录
                Lc_Mix f = new Lc_Mix
                {
                    Time = DateTime.Now,
                    Total = w.MixDisp
                };
                f.Items = new FlyData_MixItem[w.Ingredients.Count];
                for (int i = 0; i < w.Ingredients.Count; i++)
                {
                    FlyData_MixItem item = new FlyData_MixItem()
                    {
                        Disp = w.Ingredients[i].MixDisp,
                        PDisp = w.Ingredients[i].MixPDisp
                    };
                    f.Items[i] = item;
                }
                mix.Add(w, f);
            }
            mix10.Add(mix);
            if (mix10.Count > 10)
                mix10.RemoveAt(0);
        }


        //每隔1秒记录，w.MixDisp, Ingredients[i].(MixDisp,MixPDisp)，保存10s
        //当卸料时，提取2s前的数据保存！！！！
        private void WeightSystem_MBServer_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            WeighterC w = sender as WeighterC;

            if (e.PropertyName == "IsMixSave")//if (e.PropertyName == "BucketValveIsOpen")
            {
                if (w.IsMixSave)// (w.BucketValveIsOpen)
                {
                    Lc_Mix f_last = null;
                    //提取之前的记录
                    if (mix10.Count > 3)
                    {
                        f_last = mix10[mix10.Count - 3][w];
                    }


                    //记录
                    Lc_Mix f = new Lc_Mix()
                    {
                        Time = DateTime.Now,
                        Items = new FlyData_MixItem[w.Ingredients.Count]
                    };

                    if (f_last != null)// && f_last.Total > f.Total)
                    {
                        //没错,用以前的是对的！！！！
                        f.Total = Math.Round(f_last.Total, 4);
                        for (int i = 0; i < w.Ingredients.Count; i++)
                        {
                            FlyData_MixItem item = new FlyData_MixItem()
                            {
                                Disp = Math.Round(f_last.Items[i].Disp, 4),
                                PDisp = Math.Round(f_last.Items[i].PDisp, 2)
                            };
                            f.Items[i] = item;
                        }
                    }
                    else
                    {
                        f.Total = Math.Round(w.MixDisp, 4);
                        for (int i = 0; i < w.Ingredients.Count; i++)
                        {
                            FlyData_MixItem item = new FlyData_MixItem()
                            {
                                Disp = Math.Round(w.Ingredients[i].MixDisp, 4),
                                PDisp = Math.Round(w.Ingredients[i].MixPDisp, 2)
                            };
                            f.Items[i] = item;
                        }
                    }

                    int index = Items.IndexOf(w);
                    historyDb.AddMix(index, f);
                }
            }
            else if (e.PropertyName == "ClearProduction")
            {
                //当累计产量变小，肯定是被clear() 需要保存数据
                if (w.ClearProduction)
                {
                    Lc_Mix f = new Lc_Mix()
                    {
                        Time = DateTime.Now,
                        Items = new FlyData_MixItem[w.Ingredients.Count]
                    };

                    f.Total = Math.Round(w.CumulativeProduction, 4);
                    for (int i = 0; i < w.Ingredients.Count; i++)
                    {
                        FlyData_MixItem item = new FlyData_MixItem()
                        {
                            Disp = Math.Round(w.Ingredients[i].MixCum, 4),
                            PDisp = Math.Round(w.Ingredients[i].MixCumPercent, 2)
                        };
                        f.Items[i] = item;
                    }
                    int index = Items.IndexOf(w);
                    historyDb.AddRollMix(index, f);
                }
                //LastCumulativeProductions[w] = w.CumulativeProduction;
            }
        }

        void Test()
        {
            #region 测试
            Items[0].CurrentFlow = 150;
            Items[0].BinWeight = 15;
            Items[0].MixBucketWeight = 4;
            Items[0].MixPSet_1 = 90;
            Items[0].MixPSet_2 = 5;
            Items[0].MixPSet_3 = 5;
            Items[0].MixPSet_4 = 0;

            Items[1].CurrentFlow = 100;
            Items[1].BinWeight = 10;
            Items[1].MixBucketWeight = 4;
            Items[1].MixPSet_1 = 20;
            Items[1].MixPSet_2 = 70;
            Items[1].MixPSet_3 = 5;
            Items[1].MixPSet_4 = 5;

            Items[2].CurrentFlow = 50;
            Items[2].BinWeight = 5;
            Items[2].MixBucketWeight = 4;
            Items[2].MixPSet_1 = 0;
            Items[2].MixPSet_2 = 70;
            Items[2].MixPSet_3 = 30;
            Items[2].MixPSet_4 = 0;

            Items[2].IsErrorOfAdd_1 = true;
            Items[0].IsErrorOfBlender = true;

            Accessory.TotalFlow = Items.Sum((w) => { return w.CurrentFlow; });

            foreach (WeighterC w in Items)
            {
                w.FlowSetting = w.CurrentFlow;
                w.ScrewPDisp = w.CurrentFlow * 100 / Accessory.TotalFlow;
                w.ScrewPSet = w.ScrewPDisp;
                w.ScrewMotorFreq = 100 * w.CurrentFlow / 300;
                w.CumulativeProduction = w.CurrentFlow * 20;

                w.MixPDisp_1 = w.MixPSet_1;
                w.MixPDisp_2 = w.MixPSet_2;
                w.MixPDisp_3 = w.MixPSet_3;
                w.MixPDisp_4 = w.MixPSet_4;

                w.MixCum_1 = (float)(w.CumulativeProduction * w.MixPSet_1 / 100.0);
                w.MixCum_2 = (float)(w.CumulativeProduction * w.MixPSet_2 / 100.0);
                w.MixCum_3 = (float)(w.CumulativeProduction * w.MixPSet_3 / 100.0);
                w.MixCum_4 = (float)(w.CumulativeProduction * w.MixPSet_4 / 100.0);

                w.MixCumPercent_1 = w.MixPSet_1;
                w.MixCumPercent_2 = w.MixPSet_2;
                w.MixCumPercent_3 = w.MixPSet_3;
                w.MixCumPercent_4 = w.MixPSet_4;
            }


            Accessory.TotalProduction = (UInt16)Items.Sum((w) => { return w.CumulativeProduction; });
            Accessory.TotalFlowSetting = Accessory.TotalFlow;


            Random r = new Random();
            Lc_Flow last_f = null;
            int flowListSize = 10000;
            for (int j = 0; j < flowListSize; j++)
            {
                Lc_Flow f = new Lc_Flow();
                f.Time = DateTime.Now - TimeSpan.FromMinutes((flowListSize - j) * FlowInterval);

                f.Items = new FlyData_FlowItem[ItemsCnt];
                for (int i = 0; i < Items.Count; i++)
                {
                    f.Items[i] = new FlyData_FlowItem();
                    f.Items[i].Flow = (float)(Items[i].CurrentFlow + r.NextDouble() * 5 - 2.5);
                    f.Items[i].ScrewPDisp = Items[i].ScrewPDisp;
                    f.Items[i].ScrewMotorFreq = Items[i].ScrewMotorFreq;

                    //f.Items[i].CumulativeProduction = (float)(f.Items[i].Flow * FlowInterval / 60);
                    //if (last_f != null)
                    //{
                    //    f.Items[i].CumulativeProduction += last_f.Items[i].CumulativeProduction;
                    //}

                    //f.Items[i].MixCum[0] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_1 / 100.0);
                    //f.Items[i].MixCum[1] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_2 / 100.0);
                    //f.Items[i].MixCum[2] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_3 / 100.0);
                    //f.Items[i].MixCum[3] = (float)(f.Items[i].CumulativeProduction * Items[i].MixPSet_4 / 100.0);


                    //f.Items[i].MixCumPercent[0] = Items[i].MixPSet_1;
                    //f.Items[i].MixCumPercent[1] = Items[i].MixPSet_2;
                    //f.Items[i].MixCumPercent[2] = Items[i].MixPSet_3;
                    //f.Items[i].MixCumPercent[3] = Items[i].MixPSet_4;
                }
                f.Total = f.Items.Sum(i => i.Flow);
                last_f = f;
                historyDb.AddFlow(f);
            }
            #endregion
        }

        Stopwatch stopwatch_flowInterval = new Stopwatch();
        Stopwatch stopwatch_1min = new Stopwatch();
        List<Lc_Flow> lcFlows_1min = new List<Lc_Flow>();

        void OnPoll()
        {
            if (stopwatch_1min.IsRunning && stopwatch_1min.Elapsed >= TimeSpan.FromSeconds(20))
            {
                //启动了1min 且到点了
                if (lcFlows_1min.Count > 0)
                {
                    historyDb.AddFlowRange(lcFlows_1min);
                    lcFlows_1min.Clear();
                    stopwatch_1min.Restart();
                }
                else
                {
                    stopwatch_1min.Reset();
                }
            }

            if (!stopwatch_flowInterval.IsRunning)
            {
                stopwatch_flowInterval.Start();
                return;
            }
            else
            {
                if (stopwatch_flowInterval.Elapsed < TimeSpan.FromSeconds(FlowInterval))
                    return;
                else
                    stopwatch_flowInterval.Restart();
            }

            if (!PLCos.IsConnectedWithPLC)//连接断开
                return;

            if (Accessory.TotalFlow < 5)//没有生产
                return;

            //记录数据
            Lc_Flow lc_Flow = new Lc_Flow
            {
                Time = DateTime.Now,
                Total = Math.Round(Accessory.TotalFlow, 4),
                Items = new FlyData_FlowItem[Items.Count]
            };
            for (int i = 0; i < Items.Count; i++)
            {
                lc_Flow.Items[i] = new FlyData_FlowItem
                {
                    Flow = Math.Round(Items[i].CurrentFlow, 4),
                    ScrewPDisp = Math.Round(Items[i].ScrewPDisp, 2),
                    ScrewMotorFreq = Math.Round(Items[i].ScrewMotorFreq, 1)
                };
            }

            //写入数据库
            if (!stopwatch_1min.IsRunning)
            {
                //第1次立刻写入
                historyDb.AddFlow(lc_Flow);
                stopwatch_1min.Start();
            }
            else
            {
                //第2次，1分钟后才能保存
                lcFlows_1min.Add(lc_Flow);
            }

        }


        void AddConfigFile()
        {
            plcos.AddConfigFile("plcgroup.json", (plcgroup) =>
            {
                //从plcgroup 分析出 有多少层，每层的仓数
                List<int> bincnts = new List<int>();
                Regex r = new Regex(@"Items\[([0-9])\]");
                Regex r2 = new Regex("MixPSet_([1-9])");

                foreach (PLCGroup.PLCVariable var in plcgroup.Variables)
                {
                    Match m = r.Match(var.OwnerName);
                    if (m.Success)
                    {
                        int weighter_idx = int.Parse(m.Groups[1].Value);
                        if (bincnts.Count() <= weighter_idx)
                        {
                            while (bincnts.Count() <= weighter_idx)
                                bincnts.Add(0);
                        }
                        Match m2 = r2.Match(var.PropertyName);
                        if (m2.Success)
                        {
                            int bincnt = int.Parse(m2.Groups[1].Value);
                            if (bincnts[weighter_idx] < bincnt)
                                bincnts[weighter_idx] = bincnt;
                        }
                    }
                }
                BinCnts = bincnts.ToArray();
                ItemsCnt = BinCnts.Count();

                for (int i = 0; i < BinCnts.Count(); i++)
                {
                    WeighterC w = new WeighterC(((char)('A' + i)).ToString(), BinCnts[i]);
                    Items.Add(w);
                }


                //objname 转 obj
                Dictionary<string, INotifyPropertyChanged> objnames = new Dictionary<string, INotifyPropertyChanged>();

                objnames.Add("Accessory", Accessory);
                for (int i = 0; i < ItemsCnt; i++)
                    objnames.Add($"Items[{i}]", Items[i]);
                return objnames;
            });

            //这个是从测厚仪获取的，每几秒就写一次，不调试！！！！！
            plcos.IgnoreLogProperties.Add(new SenderProperty() { sender = Accessory, propertyName = "CurrentVelocitySet" });

        }

        public event PropertyChangedEventHandler PropertyChanged;

        void Save()
        {
            string json = JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
            File.WriteAllText("weight.json", json);
        }
        void Load()
        {
            if (!File.Exists("weight.json"))
                return;
            try
            {
                string json = File.ReadAllText("weight.json");
                JsonConvert.PopulateObject(json, this);
            }
            catch
            {

            }
        }


        public string[] GetSyncPropNames()
        {
            return new string[]{
                "FlowInterval",
                "ItemsCnt",
                "BinCnts"
                };
        }

        public string[] GetNoSyncPropNames()
        {
            return null;
        }
    }

}
